/*
const bool shadowcolor1Clear  = false;
*/

layout(location = 0) out vec4 data0;
layout(location = 1) out vec4 lpvImage;

#include "/lib/head.glsl"
#include "/lib/shadowconst.glsl"
#include "/lib/util/encoders.glsl"

uniform vec3 cameraPosition, previousCameraPosition;

uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

#include "/lib/voxel/store.glsl"

in vec2 uv;

#include "/lib/light/emission.glsl"

vec3 getVoxelEmission(ivec3 index) {
    ivec2 uv    = getVoxelPixel(index);

    vec4 voxel  = texelFetch(shadowcolor0, uv, 0);
    vec4 albedo = vec4(unpack2x8(voxel.x), unpack2x8(voxel.y));
    ivec2 voxelID = unpack2x8I(voxel.z);

    return getEmissionVoxel(voxelID.x, albedo.rgb, 0.0);
}
vec3 getVoxelEmission(vec4 voxel) {
    vec4 albedo = vec4(unpack2x8(voxel.x), unpack2x8(voxel.y));
    ivec2 voxelID = unpack2x8I(voxel.z);

    return getEmissionVoxel(voxelID.x, albedo.rgb, 0.0);
}

vec3 getSkylightInjection(ivec3 index) {
    if (index.y != (volumeDepth - 1)) {
        return vec3(0.0);
    } else {
        ivec2 uv    = getVoxelPixel(index);

        vec4 voxel  = texelFetch(shadowcolor0, uv, 0);

        return vec3(0.6, 0.85, 1.0) * float(voxel.a < 0.1) * tau;
    }
}

void main() {
    ivec2 pixelUV   = ivec2(floor(gl_FragCoord.xy));

    data0           = texelFetch(shadowcolor0, pixelUV, 0);

    ivec3 voxelIndex        = uvToVoxelIndex(pixelUV);

    if (uv.y < 0.5 && !outsideVoxelVolume(voxelIndex)) {
        lpvImage        = vec4(0);

        vec3 scenePos           = voxelToSceneSpace(voxelIndex) + vec3(0.5);
        vec3 velocity           = vec3(floor(cameraPosition) - floor(previousCameraPosition));
        vec3 previousScenePos   = (scenePos + velocity);
        ivec3 previousIndex     = sceneToVoxelIndexNoCam(previousScenePos);
        ivec2 previousUV        = getVoxelPixel(previousIndex);

        vec4 history            = clamp16F(texelFetch(shadowcolor1, previousUV, 0));

        if (previousUV == ivec2(-1)) history = vec4(0);

        if (clamp16F(history) != history) history = vec4(0);

        history.rgb += getVoxelEmission(data0);

        lpvImage    = clamp16F(history);
    } else {
        lpvImage    = vec4(0,0,0,0);
    }
}